from __future__ import (
    annotations,
)

from abc import (
    ABC,
)
from collections.abc import (
    Hashable,
    Iterable,
)
from typing import (
    Optional,
)


class DatabaseOperation(ABC):
    """Database Operation base class."""

    def __init__(self, *args, lock: Optional[Hashable] = None, timeout: Optional[float] = None, **kwargs):
        self.lock = lock
        self.timeout = timeout


class ComposedDatabaseOperation(DatabaseOperation):
    """Composed Database Operation class."""

    def __init__(self, operations: Iterable[DatabaseOperation], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.operations = tuple(operations)


class DatabaseOperationFactory(ABC):
    """Database Operation Factory base class."""



from .clients import (
    MockedDatabaseClient,
)
from .factories import (
    MockedLockDatabaseOperationFactory,
    MockedManagementDatabaseOperationFactory,
)
from .operations import (
    MockedDatabaseOperation,
)



from .abc import (
    DatabaseClient,
    DatabaseClientBuilder,
)
from .exceptions import (
    ConnectionException,
    DatabaseClientException,
    IntegrityException,
    ProgrammingException,
)



from abc import (
    ABC,
    abstractmethod,
)

from .operations import (
    DatabaseOperation,
    DatabaseOperationFactory,
)


class ManagementDatabaseOperationFactory(DatabaseOperationFactory, ABC):
    """Management Database Operation Factory base class."""

    @abstractmethod
    def build_create(self, database: str) -> DatabaseOperation:
        """Build the database operation to create a database.

        :param database: The new database's name.
        :return: A ``DatabaseOperation``.
        """

    @abstractmethod
    def build_delete(self, database: str) -> DatabaseOperation:
        """Build the database operation to create a database.

        :param database: The name of the database to be deleted.
        :return: A ``DatabaseOperation``.
        """



from ....database import (
    DatabaseOperation,
    ManagementDatabaseOperationFactory,
)
from ..clients import (
    MockedDatabaseClient,
)
from ..operations import (
    MockedDatabaseOperation,
)


class MockedManagementDatabaseOperationFactory(ManagementDatabaseOperationFactory):
    """For testing purposes"""

    def build_create(self, database: str) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("create")

    def build_delete(self, database: str) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("delete")


MockedDatabaseClient.set_factory(ManagementDatabaseOperationFactory, MockedManagementDatabaseOperationFactory)



from .clients import (
    ConnectionException,
    DatabaseClient,
    DatabaseClientBuilder,
    DatabaseClientException,
    IntegrityException,
    ProgrammingException,
)
from .locks import (
    DatabaseLock,
    LockDatabaseOperationFactory,
)
from .managements import (
    ManagementDatabaseOperationFactory,
)
from .mixins import (
    DatabaseMixin,
)
from .operations import (
    ComposedDatabaseOperation,
    DatabaseOperation,
    DatabaseOperationFactory,
)
from .pools import (
    DatabaseClientPool,
    DatabaseLockPool,
)



from ...exceptions import (
    MinosException,
)


class DatabaseClientException(MinosException):
    """Base exception for database client."""


class ConnectionException(DatabaseClientException):
    """Exception to be raised when database client is not able to connect to the database."""


class IntegrityException(DatabaseClientException):
    """Exception to be raised when an integrity check is not satisfied."""


class ProgrammingException(DatabaseClientException):
    """Exception to be raised when an integrity check is not satisfied."""



import logging
from asyncio import (
    sleep,
)
from collections.abc import (
    Hashable,
)
from typing import (
    Optional,
)

from aiomisc.pool import (
    ContextManager,
)

from ..config import (
    Config,
)
from ..locks import (
    LockPool,
)
from ..pools import (
    Pool,
)
from .clients import (
    ConnectionException,
    DatabaseClient,
    DatabaseClientBuilder,
)
from .locks import (
    DatabaseLock,
)

logger = logging.getLogger(__name__)


class DatabaseClientPool(Pool[DatabaseClient]):
    """Database Client Pool class."""

    def __init__(self, client_builder: DatabaseClientBuilder, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._client_builder = client_builder

    @classmethod
    def _from_config(cls, config: Config, identifier: Optional[str] = None, **kwargs):
        base_builder = config.get_database_by_name(identifier).get("client")
        if base_builder is None:
            raise ValueError(f"{base_builder!r} is not a {DatabaseClientBuilder!r} instance.")
        elif issubclass(base_builder, DatabaseClient):
            base_builder = base_builder.get_builder()
        elif issubclass(base_builder, DatabaseClientBuilder):
            base_builder = base_builder()

        client_builder = base_builder.with_name(identifier).with_config(config)

        return cls(client_builder=client_builder, **kwargs)

    async def _create_instance(self) -> Optional[DatabaseClient]:
        instance = self._client_builder.build()

        try:
            await instance.setup()
        except ConnectionException:
            await sleep(0.1)
            return None

        logger.info(f"Created {instance!r}!")
        return instance

    async def _destroy_instance(self, instance: DatabaseClient):
        if instance is None:
            return
        logger.info(f"Destroyed {instance!r}!")
        await instance.destroy()

    async def _check_instance(self, instance: Optional[DatabaseClient]) -> bool:
        if instance is None:
            return False
        return await instance.is_valid()

    async def _release_instance(self, instance: DatabaseClient) -> None:
        await instance.reset()

    @property
    def client_cls(self) -> type[DatabaseClient]:
        """Get the instance's class.

        :return: A ``type`` instance that is subclass of ``DatabaseClient``.
        """
        return self.client_builder.instance_cls

    @property
    def client_builder(self) -> DatabaseClientBuilder:
        """Get the client builder class.

        :return: A ``DatabaseClientBuilder`` instance.
        """
        return self._client_builder


class DatabaseLockPool(LockPool, DatabaseClientPool):
    """Database Lock Pool class."""

    def acquire(self, key: Hashable, *args, **kwargs) -> DatabaseLock:
        """Acquire a new lock.

        :param key: The key to be used for locking.
        :return: A ``DatabaseLock`` instance.
        """
        acquired = super(DatabaseClientPool, self).acquire()

        async def _fn_enter():
            client = await acquired.__aenter__()
            return await DatabaseLock(client, key, *args, **kwargs).__aenter__()

        async def _fn_exit(lock: DatabaseLock):
            await lock.__aexit__(None, None, None)
            await acquired.__aexit__(None, None, None)

        # noinspection PyTypeChecker
        return ContextManager(_fn_enter, _fn_exit)



from abc import (
    ABC,
    abstractmethod,
)

from ..operations import (
    DatabaseOperation,
    DatabaseOperationFactory,
)


class LockDatabaseOperationFactory(DatabaseOperationFactory, ABC):
    """Lock Database Operation Factory class."""

    @abstractmethod
    def build_acquire(self, hashed_key: int) -> DatabaseOperation:
        """Build the database operation to acquire the lock.

        :param hashed_key: The hashed key that identifies the lock.
        :return: A ``DatabaseOperation`` instance.
        """

    @abstractmethod
    def build_release(self, hashed_key: int) -> DatabaseOperation:
        """Build the database operation to release the lock.

        :param hashed_key: The hashed key that identifies the lock.
        :return: A ``DatabaseOperation`` instance.
        """



from .database import (
    MockedDatabaseClient,
    MockedDatabaseOperation,
    MockedLockDatabaseOperationFactory,
    MockedManagementDatabaseOperationFactory,
)
from .testcases import (
    DatabaseMinosTestCase,
    MinosTestCase,
)



from ....database import (
    DatabaseOperation,
    LockDatabaseOperationFactory,
)
from ..clients import (
    MockedDatabaseClient,
)
from ..operations import (
    MockedDatabaseOperation,
)


class MockedLockDatabaseOperationFactory(LockDatabaseOperationFactory):
    """For testing purposes"""

    def build_acquire(self, hashed_key: int) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("acquire")

    def build_release(self, hashed_key: int) -> DatabaseOperation:
        """For testing purposes"""
        return MockedDatabaseOperation("release")


MockedDatabaseClient.set_factory(LockDatabaseOperationFactory, MockedLockDatabaseOperationFactory)
